require 'spec_helper'

describe Solve do
  describe "ClassMethods" do
    subject { Solve }

    describe "#it" do
      it "returns nil if a solution does not exist" do
        skip
      end
    end

    describe "#it!" do
      it "raises NoSolutionError if a solution does not exist" do
        skip
      end
    end
  end
end
